/* Miraculum LZSS compressor */
/* This code is in the public domain */
/* Revision 2007-10-14 (Y-M-D) */

/* The window searching code is brute-force, dumb, and slow,
   but it should catch all compression opportunities...
*/

/* Output format is(may be partially wrong, it was written a while after writing this actual program ;)):

	(The initial window read/write offset is 0xEF)

	4 bytes - Uncompressed length, little-endian format

	followed by 1 or more chunks of the format:
	
	 1 byte - Flags(read/written most-significant-bit first, 1 = literal byte, 0 = LZSS "pair")

	 followed by the mix of 8 literal bytes and LZSS "pairs".  Literal bytes are just that, literal. :b

	 LZSS "pairs":
	
	 1 byte - Offset into the window
	 1 byte - Length.  Upper 4 bits are for the current offset, lower 4 bits are for the next offset(which will of course lack the length byte after it); 
	 	  the next offset can even be in a following chunk.

		  The 4-bit length represents lengths 2-17, mapped to actual values of 0-15(real_length = recorded_length + 2).
		  Reasoning:  Having a length of 0 would be nonsensical, and a length of 1 would result in a larger output
		  than just using a literal byte.

*/

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <errno.h>
#include <vector>
#include <string.h>

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

typedef uint8_t uint8;
typedef uint16_t uint16;
typedef uint32_t uint32;
typedef int8_t int8;
typedef int16_t int16;
typedef int32_t int32;

typedef struct
{
 char is_literal;

 union
 {
  uint8 literal_byte;
  struct
  {
   uint8 length;
   uint8 offset;
  } len_offs_pair;
 };

} CompEntry;

std::vector<CompEntry> PunkBunny;

int main(int argc, char *argv[])
{
 FILE *in_fp, *out_fp;
 uint8 *input_data;
 uint32 input_len;

 if(argc < 3)
 {
  printf("Usage: %s input.bin output.bin\n", argv[0]);
  return(-1);
 }
 in_fp = fopen(argv[1], "rb");
 if(!in_fp)
 {
  printf("Error opening %s: %s\n", argv[1], strerror(errno));
  return(-1);
 }

 fseek(in_fp, 0, SEEK_END);
 input_len = ftell(in_fp);
 fseek(in_fp, 0, SEEK_SET);

 input_data = (uint8 *)malloc(input_len);
 if(fread(input_data, 1, input_len, in_fp) != input_len)
 {
  printf("Error reading input data.\n");
  fclose(in_fp);
  return(-1);
 }

 out_fp = fopen(argv[2], "wb");
 if(!out_fp)
 {
  printf("Error opening %s: %s\n", argv[2], strerror(errno));
  fclose(in_fp);
  return(-1);
 }

 uint8 window[0x100];
 uint8 window_wr_index = 0xEF;

 memset(window, 0, sizeof(window));

 unsigned int i = 0;

 while(i < input_len)
 {
  int ll_max = (input_len - i) > 17 ? 17 : input_len - i;
  int ll;
  int wsp;

  if(ll_max == 1)
  {
   ll = 1;
   goto BreakOut;
  }

  for(ll = ll_max; ll >= 2; ll--)
  {
   for(wsp = 0; wsp < 0x100; wsp++)
   {
    bool match = TRUE;
    uint8 window_simulate[0x100];
    uint32 window_simulate_wr_index = window_wr_index;

    memcpy(window_simulate, window, 0x100);

    for(int lala = 0; lala < ll; lala++)
    {
     if(window_simulate[(wsp + lala) & 0xFF] != input_data[i + lala])
     {
      match = FALSE;
      break;
     }
     window_simulate[window_simulate_wr_index] = input_data[i + lala];
     window_simulate_wr_index++;
    }

    // Hurray!
    if(match)
     goto BreakOut;
   }
  }

  BreakOut:

  if(ll == 1) // hmph, literal!
  {
   CompEntry kitty;
   kitty.is_literal = TRUE;
   kitty.literal_byte = input_data[i];
   PunkBunny.push_back(kitty);
  }
  else
  {
   CompEntry kitty;
   kitty.is_literal = FALSE;
   kitty.len_offs_pair.length = ll;
   kitty.len_offs_pair.offset = wsp;
   PunkBunny.push_back(kitty);
  }

  // Update the sliding window...OF DOOM
  for(int cowcat = 0; cowcat < ll; cowcat++)
  {
   window[window_wr_index] = input_data[i];
   window_wr_index++;
   i++;
  }
 }

 //fwrite(&input_len, 1, 4, out_fp);
 fputc(input_len >> 0, out_fp);
 fputc(input_len >> 8, out_fp);
 fputc(input_len >> 16, out_fp);
 fputc(input_len >> 24, out_fp);

 int len_flip = 0;

 // Now, encode and write out to the output!
 for(int i = 0; i < PunkBunny.size(); i += 8)
 {
  int zmax = (PunkBunny.size() - i) > 8 ? 8 : PunkBunny.size() - i;
  uint8 flags_byte = 0;

  for(int z = 0; z < zmax; z++)
   if(PunkBunny[i + z].is_literal)
   {
    flags_byte |= 1 << (7 - z);
   }
  //printf("\nFlags: 0x%02x\n", flags_byte);
  fputc(flags_byte, out_fp);

  for(int z = 0; z < zmax; z++)
  {
   //if(PunkBunny[i + z].is_literal)
   // printf("%d, Literal: 0x%02x\n", i + z, PunkBunny[i + z].literal_byte);
   //else
   // printf("%d, Offset: 0x%02x, Length: %d\n", i + z, PunkBunny[i + z].len_offs_pair.offset, PunkBunny[i + z].len_offs_pair.length);

   if(PunkBunny[i + z].is_literal)
   {
    fputc(PunkBunny[i + z].literal_byte, out_fp);
    //puts(" wrote literal");
   }
   else
   {
    fputc(PunkBunny[i + z].len_offs_pair.offset, out_fp);
    //puts(" wrote offset");

    if(!len_flip)
    {
     uint8 len_byte = (PunkBunny[i + z].len_offs_pair.length - 2) << 4;

     for(int zs = i + z + 1; zs < PunkBunny.size(); zs++)
      if(!PunkBunny[zs].is_literal)
      {
       len_byte |= PunkBunny[zs].len_offs_pair.length - 2;
       break;
      }

     len_flip = 1;
     fputc(len_byte, out_fp);
     //puts(" wrote length");
    }
    else
     len_flip = 0;
   }
  }
 }

 printf("Input:\t%ld bytes\nOutput:\t%ld bytes\n\n", (long)input_len, (long)ftell(out_fp));

 fclose(in_fp);
 fclose(out_fp);
}

